from libc.stdio cimport FILE
from libc.stdint cimport uint32_t

from libeasel cimport esl_pos_t


cdef extern from "esl_keyhash.h" nogil:

    ctypedef struct ESL_KEYHASH:
        int* hashtable
        uint32_t hashsize
        int* key_offset
        int* nxt
        int nkeys
        int kalloc

        char* smem
        int salloc
        int sn

    ESL_KEYHASH *esl_keyhash_Create()
    ESL_KEYHASH *esl_keyhash_CreateCustom(uint32_t hashsize, int kalloc, int salloc)
    ESL_KEYHASH *esl_keyhash_Clone(const ESL_KEYHASH *kh)
    char*        esl_keyhash_Get(const ESL_KEYHASH *kh, int idx)
    int          esl_keyhash_GetNumber(const ESL_KEYHASH *kh)
    size_t       esl_keyhash_Sizeof(const ESL_KEYHASH *kh)
    int          esl_keyhash_Reuse(ESL_KEYHASH *kh)
    void         esl_keyhash_Destroy(ESL_KEYHASH *kh)
    void         esl_keyhash_Dump(FILE *fp, const ESL_KEYHASH *kh)

    int  esl_keyhash_Store (      ESL_KEYHASH *kh, const char *key, esl_pos_t n, int *ret_index) except -1
    int  esl_keyhash_Lookup(const ESL_KEYHASH *kh, const char *key, esl_pos_t n, int *ret_index)



cdef extern from "esl_huffman.h" nogil:

    cdef size_t eslHUFFMAN_MAXCODE

    ctypedef huffman_s ESL_HUFFMAN
    cdef struct huffman_s:
        int* len
        uint32_t* code
        int K
        int* sorted_at
        int Ku
        int* dt_len
        uint32_t* dt_lcode
        int* dt_rank
        int D
        int Lmax


    int  esl_huffman_Build(const float *fq, int K, ESL_HUFFMAN **ret_hc)
    void esl_huffman_Destroy(ESL_HUFFMAN *hc)

    int  esl_huffman_Encode(const ESL_HUFFMAN *hc, const char     *T, int n,  uint32_t **ret_X, int *ret_nb)
    int  esl_huffman_Decode(const ESL_HUFFMAN *hc, const uint32_t *X, int nb, char     **ret_T, int *ret_n)

    int  esl_huffman_Dump(FILE *fp, ESL_HUFFMAN *hc)



from libc.stdint cimport int32_t, int64_t, uint32_t
from libc.stdio cimport FILE
from posix.types cimport off_t

from libeasel cimport ESL_DSQ
from libeasel.msa cimport ESL_MSA
from libeasel.msafile cimport ESL_MSAFILE
from libeasel.sq cimport ESL_SQ
from libeasel.sqio cimport esl_sqio_s
from libeasel.ssi cimport ESL_SSI


cdef extern from "easel.h":
    DEF eslERRBUFSIZE = 128


cdef extern from "esl_sqio.h" nogil:
    DEF MAX_RESIDUE_COUNT = 1024*1024

    ctypedef esl_sqascii_s ESL_SQASCII_DATA
    cdef struct esl_sqascii_s:
        FILE *fp
        char[eslERRBUFSIZE]  errbuf

        bint   do_gzip
        bint   do_stdin
        bint   do_buffer

        char    *mem
        int      allocm
        int      mn
        int      mpos
        off_t    moff
        bint      is_recording

        char    *buf
        off_t    boff
        int      balloc
        int      nc
        int      bpos
        int64_t  L
        int64_t  linenumber
        off_t    bookmark_offset
        int64_t  bookmark_linenum

        bint   is_linebased
        bint   eof_is_ok
        int  (*parse_header)(esl_sqio_s *, ESL_SQ *sq)
        int  (*skip_header) (esl_sqio_s *, ESL_SQ *sq)
        int  (*parse_end)   (esl_sqio_s *, ESL_SQ *sq)

        ESL_MSAFILE  *afp
        ESL_MSA      *msa
        int           idx

        char    *ssifile
        int      rpl
        int      bpl
        int      currpl
        int      curbpl
        int      prvrpl
        int      prvbpl
        ESL_SSI *ssi

    int  esl_sqascii_Open(char *seqfile, int format, esl_sqio_s *sqfp)
    int  esl_sqascii_WriteFasta(FILE *fp, ESL_SQ *s, int update)
    int  esl_sqascii_Parse(char *buf, int size, ESL_SQ *s, int format)



from libc.stdint cimport uint8_t, int16_t
from libc.stdio cimport FILE
from posix.types cimport off_t

from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.random cimport ESL_RANDOMNESS
from libhmmer.p7_bg cimport P7_BG
from libhmmer.p7_hmm cimport P7_HMM
from libhmmer.p7_profile cimport P7_PROFILE


cdef extern from "p7_config.h" nogil:
    DEF p7_MAXABET = 20


cdef extern from "hmmer.h" nogil:
    DEF p7_NEVPARAM = 6
    DEF p7_NCUTOFFS = 6
    DEF p7_NOFFSETS = 3


cdef extern from "impl_vmx/impl_vmx.h" nogil:

    cdef int p7O_NQB(int)
    cdef int p7O_NQW(int)
    cdef int p7O_NQF(int)

    DEF p7O_NXSTATES = 4
    cdef enum p7o_xstates_e:
        p7O_E = 0
        p7O_N = 1
        p7O_J = 2
        p7O_C = 3

    DEF p7O_NXTRANS = 2
    cdef enum p7o_xtransitions_e:
        p7O_MOVE = 0
        p7O_LOOP = 1

    DEF p7O_NTRANS = 8
    cdef enum p7o_tsc_e:
        p7O_BM   = 0
        p7O_MM   = 1
        p7O_IM   = 2
        p7O_DM   = 3
        p7O_MD   = 4
        p7O_MI   = 5
        p7O_II   = 6
        p7O_DD   = 7

    ctypedef p7_oprofile_s P7_OPROFILE
    cdef struct p7_oprofile_s:
        # vector unsigned char** rbv
        uint8_t   tbm_b
        uint8_t   tec_b
        uint8_t   tjb_b
        float     scale_b
        uint8_t   base_b
        uint8_t   bias_b

        # vector signed short **rwv
        # vector signed short  *twv
        int16_t[p7O_NXSTATES][p7O_NXTRANS]   xw
        float     scale_w
        int16_t   base_w
        int16_t   ddbound_w
        float     ncj_roundoff

        # vector float **rfv
        # vector float  *tfv
        float[p7O_NXSTATES][p7O_NXTRANS]    xf

        # vector unsigned char  *rbv_mem
        # vector signed short   *rwv_mem
        # vector signed short   *twv_mem
        # vector float          *tfv_mem
        # vector float          *rfv_mem

        off_t[p7_NOFFSETS]  offs
        off_t  roff
        off_t  eoff

        char  *name
        char  *acc
        char  *desc
        char  *rf
        char  *mm
        char  *cs
        char  *consensus
        float[p7_NEVPARAM]  evparam
        float[p7_NCUTOFFS]  cutoff
        float[p7_MAXABET]  compo
        const ESL_ALPHABET *abc

        int L
        int M
        int max_length
        int allocM
        int allocQ4
        int allocQ8
        int allocQ16
        int mode
        float nj

        int clone

    ctypedef struct P7_OM_BLOCK:
        int count
        int listSize
        P7_OPROFILE** list


    P7_OPROFILE *p7_oprofile_Create(int M, const ESL_ALPHABET *abc)
    int          p7_oprofile_IsLocal(const P7_OPROFILE *om)
    void         p7_oprofile_Destroy(P7_OPROFILE *om)
    size_t       p7_oprofile_Sizeof(P7_OPROFILE *om)
    P7_OPROFILE *p7_oprofile_Copy(P7_OPROFILE *om)
    P7_OPROFILE *p7_oprofile_Clone(const P7_OPROFILE *om)
    int          p7_oprofile_UpdateFwdEmissionScores(P7_OPROFILE *om, P7_BG *bg, float *fwd_emissions, float *sc_arr)
    int          p7_oprofile_UpdateVitEmissionScores(P7_OPROFILE *om, P7_BG *bg, float *fwd_emissions, float *sc_arr)
    int          p7_oprofile_UpdateMSVEmissionScores(P7_OPROFILE *om, P7_BG *bg, float *fwd_emissions, float *sc_arr)

    int          p7_oprofile_Convert(const P7_PROFILE *gm, P7_OPROFILE *om)
    int          p7_oprofile_ReconfigLength    (P7_OPROFILE *om, int L)
    int          p7_oprofile_ReconfigMSVLength (P7_OPROFILE *om, int L)
    int          p7_oprofile_ReconfigRestLength(P7_OPROFILE *om, int L)
    int          p7_oprofile_ReconfigMultihit  (P7_OPROFILE *om, int L)
    int          p7_oprofile_ReconfigUnihit    (P7_OPROFILE *om, int L)

    int          p7_oprofile_Dump(FILE *fp, const P7_OPROFILE *om)
    int          p7_oprofile_Sample(ESL_RANDOMNESS *r, const ESL_ALPHABET *abc, const P7_BG *bg, int M, int L,
    				       P7_HMM **opt_hmm, P7_PROFILE **opt_gm, P7_OPROFILE **ret_om)
    int          p7_oprofile_Compare(const P7_OPROFILE *om1, const P7_OPROFILE *om2, float tol, char *errmsg)
    int          p7_profile_SameAsMF(const P7_OPROFILE *om, P7_PROFILE *gm)
    int          p7_profile_SameAsVF(const P7_OPROFILE *om, P7_PROFILE *gm)

    int          p7_oprofile_GetFwdTransitionArray(const P7_OPROFILE *om, int type, float *arr )
    int          p7_oprofile_GetSSVEmissionScoreArray(const P7_OPROFILE *om, uint8_t *arr )
    int          p7_oprofile_GetFwdEmissionScoreArray(const P7_OPROFILE *om, float *arr )
    int          p7_oprofile_GetFwdEmissionArray(const P7_OPROFILE *om, P7_BG *bg, float *arr )



cdef extern from "x86intrin.h":

    ctypedef struct __m256i:
        pass

    ctypedef struct __m256:
        pass


cdef extern from "esl_avx512.h" nogil:

    void esl_avx512_dump_512i_hex8(__m512i v)
    uint8_t esl_avx512_hmax_epu8(__m512i a)
    int8_t esl_avx512_hmax_epi8(__m512i a)
    int16_t esl_avx512_hmax_epi16(__m512i a)
    void esl_avx512_hsum_ps(__m512 a, float *ret_sum)
    __m512i esl_avx512_rightshift_int8(__m512i v, __m512i neginfmask)
    __m512i esl_avx512_rightshift_int16(__m512i v, __m512i neginfmask)
    __m512 esl_avx512_rightshiftz_float(__m512 v)
    __m512 esl_avx512_leftshiftz_float(__m512 v)



from libc.stdint cimport uint8_t, int16_t
from libc.stdio cimport FILE
from posix.types cimport off_t

from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.random cimport ESL_RANDOMNESS
from libhmmer.p7_bg cimport P7_BG
from libhmmer.p7_hmm cimport P7_HMM
from libhmmer.p7_profile cimport P7_PROFILE


cdef extern from "p7_config.h" nogil:
    DEF p7_MAXABET = 20


cdef extern from "hmmer.h" nogil:
    DEF p7_NEVPARAM = 6
    DEF p7_NCUTOFFS = 6
    DEF p7_NOFFSETS = 3


cdef extern from "emmintrin.h":

    ctypedef struct __m128i:
        pass

    ctypedef struct __m128:
        pass


cdef extern from "impl_sse/impl_sse.h" nogil:

    cdef int p7O_NQB(int)
    cdef int p7O_NQW(int)
    cdef int p7O_NQF(int)

    DEF p7O_NXSTATES = 4
    cdef enum p7o_xstates_e:
        p7O_E  = 0
        p7O_N  = 1
        p7O_J  = 2
        p7O_C  = 3

    DEF p7O_NXTRANS = 2
    cdef enum p7o_xtransitions_e:
        p7O_MOVE = 0
        p7O_LOOP = 1

    DEF p7O_NTRANS = 8
    cdef enum p7o_tsc_e:
        p7O_BM = 0
        p7O_MM = 1
        p7O_IM = 2
        p7O_DM = 3
        p7O_MD = 4
        p7O_MI = 5
        p7O_II = 6
        p7O_DD = 7


    ctypedef p7_oprofile_s P7_OPROFILE
    cdef struct p7_oprofile_s:
        __m128i** rbv
        __m128i** sbv
        uint8_t tbm_b
        uint8_t tec_b
        uint8_t tjb_b
        float scale_b
        uint8_t base_b
        uint8_t bias_b

        __m128i** rwv
        __m128i*  twv
        int16_t[p7O_NXSTATES][p7O_NXTRANS] xw
        float scale_w
        int16_t base_w
        int16_t ddbound_w
        float ncj_roundoff

        __m128** rfv
        __m128* tfv
        int16_t[p7O_NXSTATES][p7O_NXTRANS] xf

        __m128i  *rbv_mem
        __m128i  *sbv_mem
        __m128i  *rwv_mem
        __m128i  *twv_mem
        __m128   *tfv_mem
        __m128   *rfv_mem

        off_t[p7_NOFFSETS]  offs
        off_t  roff
        off_t  eoff

        char* name
        char* acc
        char* desc
        char* rf
        char* mm
        char* cs
        char* consensus
        float[p7_NEVPARAM] evparam
        float[p7_NCUTOFFS] cutoff
        float[p7_MAXABET] compo
        const ESL_ALPHABET* abc

        int L
        int M
        int max_length
        int allocM
        int allocQ4
        int allocQ8
        int allocQ16
        int mode
        float nj

        int clone


    P7_OPROFILE *p7_oprofile_Create(int M, const ESL_ALPHABET *abc)
    int          p7_oprofile_IsLocal(const P7_OPROFILE *om)
    void         p7_oprofile_Destroy(P7_OPROFILE *om)
    size_t       p7_oprofile_Sizeof(P7_OPROFILE *om)
    P7_OPROFILE *p7_oprofile_Copy(P7_OPROFILE *om)
    P7_OPROFILE *p7_oprofile_Clone(const P7_OPROFILE *om)
    int          p7_oprofile_UpdateFwdEmissionScores(P7_OPROFILE *om, P7_BG *bg, float *fwd_emissions, float *sc_arr)
    int          p7_oprofile_UpdateVitEmissionScores(P7_OPROFILE *om, P7_BG *bg, float *fwd_emissions, float *sc_arr)
    int          p7_oprofile_UpdateMSVEmissionScores(P7_OPROFILE *om, P7_BG *bg, float *fwd_emissions, float *sc_arr)

    int          p7_oprofile_Convert(const P7_PROFILE *gm, P7_OPROFILE *om)
    int          p7_oprofile_ReconfigLength    (P7_OPROFILE *om, int L)
    int          p7_oprofile_ReconfigMSVLength (P7_OPROFILE *om, int L)
    int          p7_oprofile_ReconfigRestLength(P7_OPROFILE *om, int L)
    int          p7_oprofile_ReconfigMultihit  (P7_OPROFILE *om, int L)
    int          p7_oprofile_ReconfigUnihit    (P7_OPROFILE *om, int L)

    int          p7_oprofile_Dump(FILE *fp, const P7_OPROFILE *om)
    int          p7_oprofile_Sample(ESL_RANDOMNESS *r, const ESL_ALPHABET *abc, const P7_BG *bg, int M, int L,
                   P7_HMM **opt_hmm, P7_PROFILE **opt_gm, P7_OPROFILE **ret_om)
    int          p7_oprofile_Compare(const P7_OPROFILE *om1, const P7_OPROFILE *om2, float tol, char *errmsg)
    int          p7_profile_SameAsMF(const P7_OPROFILE *om, P7_PROFILE *gm)
    int          p7_profile_SameAsVF(const P7_OPROFILE *om, P7_PROFILE *gm)

    int          p7_oprofile_GetFwdTransitionArray(const P7_OPROFILE *om, int type, float *arr )
    int          p7_oprofile_GetSSVEmissionScoreArray(const P7_OPROFILE *om, uint8_t *arr )
    int          p7_oprofile_GetFwdEmissionScoreArray(const P7_OPROFILE *om, float *arr )
    int          p7_oprofile_GetFwdEmissionArray(const P7_OPROFILE *om, P7_BG *bg, float *arr )



cdef extern from "esl_histogram.h" nogil:

    cdef enum esl_dataset:
        COMPLETE
        VIRTUAL_CENSORED
        TRUE_CENSORED

    ctypedef struct ESL_HISTOGRAM:
        uint64_t* obs
        int nb
        double w
        double bmin
        double bmax
        int imin
        int imax

        double xmin
        double xmax
        uint64_t n
        double* x
        uint64_t nalloc

        double phi
        int cmin
        uint64_t z
        uint64_t Nc
        uint64_t No

        double* expect
        int emin
        double tailbase
        double tailmass

        bint is_full
        bint is_done
        bint is_sorted
        bint is_tailfit
        bint is_rounded

        esl_dataset dataset_is


    # Creating/destroying histograms and collecting data
    ESL_HISTOGRAM *esl_histogram_Create    (double bmin, double bmax, double w);
    ESL_HISTOGRAM *esl_histogram_CreateFull(double bmin, double bmax, double w);
    void           esl_histogram_Destroy  (ESL_HISTOGRAM *h);
    int            esl_histogram_Score2Bin(ESL_HISTOGRAM *h, double x, int *ret_b);
    int            esl_histogram_Add      (ESL_HISTOGRAM *h, double x);

    # Declarations about the binned data before parameter fitting
    int esl_histogram_DeclareCensoring(ESL_HISTOGRAM *h, int z, double phi);
    int esl_histogram_DeclareRounding (ESL_HISTOGRAM *h);
    int esl_histogram_SetTail         (ESL_HISTOGRAM *h, double phi,
    					  double *ret_newmass);
    int esl_histogram_SetTailByMass   (ESL_HISTOGRAM *h, double pmass,
    					  double *ret_newmass);

    # Accessing data samples in a full histogram:
    int esl_histogram_GetRank(ESL_HISTOGRAM *h, int rank, double *ret_x);
    int esl_histogram_GetData(ESL_HISTOGRAM *h, double **ret_x, int *ret_n);
    int esl_histogram_GetTail(ESL_HISTOGRAM *h, double phi, double **ret_x,
    				 int *ret_n, int *ret_z);
    int esl_histogram_GetTailByMass(ESL_HISTOGRAM *h, double pmass,
    				       double **ret_x, int *ret_n, int *ret_z);


    # Setting expected binned counts:
    int esl_histogram_SetExpect(ESL_HISTOGRAM *h,
    				   double (*cdf)(double x, void *params),
    				   void *params);
    int esl_histogram_SetExpectedTail(ESL_HISTOGRAM *h, double base_val,
    					 double pmass,
    					 double (*cdf)(double x, void *params),
    					 void *params);

    # Output/display of binned data:
    int esl_histogram_Write       (FILE *fp, ESL_HISTOGRAM *h);
    int esl_histogram_Plot        (FILE *fp, ESL_HISTOGRAM *h);
    int esl_histogram_PlotSurvival(FILE *fp, ESL_HISTOGRAM *h);
    int esl_histogram_PlotQQ      (FILE *fp, ESL_HISTOGRAM *h,
    				      double (*invcdf)(double, void *), void *params);

    # Goodness of fit testing
    int esl_histogram_Goodness(ESL_HISTOGRAM *h, int nfitted,
    				  int *ret_nbins,
    				  double *ret_G,  double *ret_Gp,
    				  double *ret_X2, double *ret_X2p);



# coding: utf-8
"""Cython bindings and Python interface to HMMER3.

HMMER is a biological sequence analysis tool that uses profile hidden Markov
models to search for sequence homologs. HMMER3 is maintained by members of the
the `Eddy/Rivas Laboratory <http://eddylab.org/>`_ at Harvard University.

``pyhmmer`` is a module, implemented using the `Cython <https://cython.org/>`_
language, that provides bindings to HMMER3. It directly interacts with the
HMMER internals, which has several advantages over CLI wrappers like
`hmmer-py <https://pypi.org/project/hmmer/>`_.

"""

import collections.abc as _collections_abc
import contextlib as _contextlib
import os as _os

from . import errors
from . import easel
from . import plan7

from .hmmer import hmmalign, hmmsearch, hmmpress, nhmmer, phmmer


__author__ = "Martin Larralde <martin.larralde@embl.de>"
__license__ = "MIT"
__version__ = "0.4.11"
__all__ = [
    errors.__name__,
    easel.__name__,
    plan7.__name__,
    hmmalign.__name__,
    hmmsearch.__name__,
    hmmpress.__name__,
    phmmer.__name__,
    nhmmer.__name__,
]

# Small addition to the docstring: we want to show a link redirecting to the
# rendered version of the documentation, but this can only work when Python
# is running with docstrings enabled
if __doc__ is not None:
    __doc__ += """See Also:
    An online rendered version of the documentation for this version of the
    library on `Read The Docs <https://pyhmmer.readthedocs.io/en/v{}/>`_.

    """.format(__version__)

# Register collections using the `collections.abc` module (this is probably
# not required with later versions of Python)
_collections_abc.Iterator.register(easel.SequenceFile)
_collections_abc.Iterator.register(plan7.HMMFile)
_collections_abc.Mapping.register(easel.KeyHash)
_collections_abc.Sized.register(plan7.Alignment)
_collections_abc.Sequence.register(easel.Bitfield)
_collections_abc.Sequence.register(plan7.Domains)
_collections_abc.Sequence.register(plan7.TopHits)

if hasattr(_contextlib, "AbstractContextManager"):
    _contextlib.AbstractContextManager.register(easel.SequenceFile)
    _contextlib.AbstractContextManager.register(easel.SSIReader)
    _contextlib.AbstractContextManager.register(easel.SSIWriter)
    _contextlib.AbstractContextManager.register(plan7.HMMFile)



#! /usr/bin/env python3

# Usage:
#    techtree.py <depfile>
#
# Where input <depfile> is created in Easel directory by clang preprocessing:
#     for cfile in esl_*.c easel.c; do clang -MM -c -I. -DeslENABLE_AVX -DeslENABLE_AVX512 -DeslENABLE_VMX -DeslENABLE_NEON -DeslENABLE_SSE -DHAVE_MPI ${cfile} >> easel.dep; done 
#
# (Some harmless errors will come out on stderr about NEON and VMX instructions not being enabled; ignore them.)
#
# This input file consists of lists of dependencies, possibly wrapped to multiple lines,
# as in:
#
#    esl_cpu.o: esl_cpu.c esl_config.h easel.h esl_cpu.h
#    esl_dirichlet.o: esl_dirichlet.c esl_config.h easel.h esl_fileparser.h \
#      esl_minimizer.h esl_random.h esl_stats.h esl_vectorops.h \
#      esl_dirichlet.h
#    esl_distance.o: esl_distance.c esl_config.h easel.h esl_alphabet.h \
#      esl_dmatrix.h esl_random.h esl_distance.h
#
# For each module, we extract its name ("esl_distance.o" => "distance"). We
# ignore the dependency on a module's own .c and .h file, and on esl_config.h.
#
# Four global data structures describe the proposed layout of the tech tree,
# as shown in the tech tree figure:
# <easelgrps> specifies an order that the groups organize into:
#   base, a layer of fundamental stuff, sequence handling,
#   statistics, misc stuff off the base, and MPI over all.
# <easelmods> specifies how modules are assigned to groups;
#    <easelmods[g]> is an ordered list of modules that belong to group <g>
# <easelparents> specifies dependencies at group level:
#    <easelparents[g]> is a set of groups that group <g> depends on.
# <easeladd> specifies additional dependencies within a group:
#    <easeladd[m]> is a set of modules that <m> depends on within m's own group.
#
# The script validates this proposed layout, and checks that the actual dependencies
# in Easel are consistent with it.
#
# The output is in two sections: the actual dependencies of each Easel module
# (so you can consider improved layouts), followed by a list of any actual
# dependencies that aren't satisfied by the current layout.
#

import sys
import re


easelgrps = [
    'BASE',
    'BIOLOGICAL_SEQUENCES', 'NUMERICAL_METHODS', 'ALGORITHMS', 'FILE_INPUT',
    'ADVANCED_SEQUENCES',   'MULTIPLE_ALIGNMENTS', 'MULTIPLE_ALIGNMENT_FILES', 'SEQUENCE_FILES',
    'STATISTICAL_DISTRIBUTIONS', 'MIXTURE_DISTRIBUTIONS',
    'COMMANDLINE', 'BENCHMARKS', 'SIMD_VECTORS', 'MULTITHREADING',
    'MPI'
    ]

easelmods = {
    'BASE' :                      [ 'easel', 'mem', 'random', 'regexp', 'stack', 'vectorops' ],
    'ALGORITHMS' :                [ 'arr2', 'arr3', 'bitfield', 'cluster', 'dmatrix', 'heap', 'keyhash', 'matrixops', 'quicksort', 'red_black', 'varint', 'huffman', 'graph', 'tree' ],
    'NUMERICAL_METHODS' :         [ 'rootfinder', 'minimizer', 'rand64' ],
    'BIOLOGICAL_SEQUENCES' :      [ 'alphabet', 'composition', 'hmm' ],
    'FILE_INPUT' :                [ 'buffer', 'fileparser', 'recorder', 'ssi', 'json' ],
    'COMMANDLINE' :               [ 'getopts', 'subcmd' ],
    'BENCHMARKS' :                [ 'stopwatch' ],
    'STATISTICAL_DISTRIBUTIONS' : [ 'stats', 'normal', 'histogram', 'exponential', 'gamma', 'gev', 'gumbel', 'stretchexp', 'weibull' ],
    'MIXTURE_DISTRIBUTIONS':      [ 'dirichlet', 'mixgev', 'hyperexp', 'mixdchlet' ],
    'ADVANCED_SEQUENCES' :        [ 'distance', 'wuss', 'paml', 'randomseq', 'ratematrix', 'scorematrix', 'swat' ],
    'MULTIPLE_ALIGNMENTS' :       [ 'msa', 'msacluster', 'msashuffle', 'msaweight' ],
    'MULTIPLE_ALIGNMENT_FILES' :  [ 'msafile', 'msafile2', 'msafile_a2m', 'msafile_afa', 'msafile_clustal', 'msafile_phylip', 'msafile_psiblast', 'msafile_selex', 'msafile_stockholm' ],
    'SEQUENCE_FILES' :            [ 'sq', 'sqio_ascii', 'sqio_ncbi', 'sqio', 'dsqdata', 'gencode' ],
    'SIMD_VECTORS' :              [ 'alloc', 'cpu', 'sse', 'avx', 'avx512', 'neon', 'vmx' ],
    'MULTITHREADING' :            [ 'threads', 'workqueue' ],
    'MPI' :                       [ 'mpi' ],
    }

easelparents = {
    'BASE' :                      {},
    'BIOLOGICAL_SEQUENCES' :      { 'BASE' },
    'NUMERICAL_METHODS' :         { 'BASE' },
    'ALGORITHMS':                 { 'BASE' },
    'FILE_INPUT':                 { 'BASE' },
    'COMMANDLINE':                { 'BASE' },
    'BENCHMARKS':                 { 'BASE' },
    'SIMD_VECTORS':               { 'BASE' },
    'MULTITHREADING':             { 'BASE' },
    'ADVANCED_SEQUENCES':         { 'BIOLOGICAL_SEQUENCES', 'NUMERICAL_METHODS', 'ALGORITHMS', 'FILE_INPUT' },
    'STATISTICAL_DISTRIBUTIONS' : { 'NUMERICAL_METHODS' },
    'MIXTURE_DISTRIBUTIONS' :     { 'STATISTICAL_DISTRIBUTIONS', 'ALGORITHMS', 'FILE_INPUT' },
    'MULTIPLE_ALIGNMENTS' :       { 'ADVANCED_SEQUENCES' },
    'MULTIPLE_ALIGNMENT_FILES' :  { 'MULTIPLE_ALIGNMENTS' },
    'SEQUENCE_FILES' :            { 'MULTIPLE_ALIGNMENT_FILES' },
    'MPI':                        { 'BENCHMARKS', 'SEQUENCE_FILES' }
    }
    
easeladd = {
# BASE
    'easel'             : set(),
    'mem'               : { 'easel' },
    'random'            : { 'easel' },
    'regexp'            : { 'easel' },
    'stack'             : { 'easel', 'random' },
    'vectorops'         : { 'easel', 'random' },
# ALGORITHMS
    'arr2'              : set(),
    'arr3'              : set(),
    'bitfield'          : set(),
    'cluster'           : set(),
    'dmatrix'           : set(),
    'heap'              : set(),
    'keyhash'           : set(),
    'matrixops'         : set(),
    'quicksort'         : set(),
    'red_black'         : set(),
    'varint'            : set(),
    'huffman'           : { 'quicksort' },
    'graph'             : { 'matrixops' },
    'tree'              : { 'arr2', 'dmatrix', 'stack'},
# NUMERICAL_METHODS
    'rootfinder'        : set(),
    'minimizer'         : set(),
    'rand64'            : set(),
# BIOLOGICAL_SEQUENCES
    'alphabet'          : set(),
    'composition'       : set(),
    'hmm'               : { 'alphabet' },
# FILE_INPUT
    'buffer'            : set(),
    'fileparser'        : set(),
    'recorder'          : set(),
    'ssi'               : set(),
    'json'              : { 'buffer' },
# COMMANDLINE
    'getopts'           : set(),
    'subcmd'            : { 'getopts' },
# BENCHMARKS
    'stopwatch'         : set(),
# STATISTICAL_DISTRIBUTIONS
    'stats'             : set(),
    'normal'            : { 'stats' },
    'histogram'         : { 'stats' },
    'exponential'       : { 'stats', 'histogram' },
    'gamma'             : { 'stats', 'histogram' },
    'gev'               : { 'stats' },
    'gumbel'            : { 'stats' },
    'stretchexp'        : { 'stats', 'histogram' },
    'weibull'           : { 'stats', 'histogram' },
# MIXTURE_DISTRIBUTIONS
    'dirichlet'         : set(),
    'hyperexp'          : set(),
    'mixdchlet'         : { 'dirichlet' },
    'mixgev'            : { 'dirichlet' },
# ADVANCED_SEQUENCES
    'distance'          : set(),
    'wuss'              : set(),
    'paml'              : set(),
    'randomseq'         : set(),
    'ratematrix'        : set(),
    'scorematrix'       : { 'ratematrix'  },
    'swat'              : { 'scorematrix' },       
# MULTIPLE_ALIGNMENTS
    'msa'               : set(),
    'msacluster'        : { 'msa' },
    'msashuffle'        : { 'msa' },
    'msaweight'         : { 'msa', 'msacluster' },
# MULTIPLE_ALIGNMENT_FILES
    'msafile'           : set(),
    'msafile2'          : set(),
    'msafile_a2m'       : { 'msafile' },
    'msafile_afa'       : { 'msafile' },
    'msafile_clustal'   : { 'msafile' },
    'msafile_phylip'    : { 'msafile' },
    'msafile_psiblast'  : { 'msafile' },
    'msafile_selex'     : { 'msafile' },
    'msafile_stockholm' : { 'msafile' },
# SEQUENCE_FILES
    'sq'                : set(),
    'sqio'              : { 'sq' },
    'sqio_ascii'        : { 'sq', 'sqio' },
    'sqio_ncbi'         : { 'sq', 'sqio' },
    'dsqdata'           : { 'sq', 'sqio', 'sqio_ascii' },
    'gencode'           : { 'sq', 'sqio', 'sqio_ascii', 'sqio_ncbi' },
# SIMD_VECTORS
    'alloc'             : set(),
    'cpu'               : set(),
    'sse'               : set(),
    'avx'               : set(),
    'avx512'            : set(),
    'neon'              : set(),
    'vmx'               : set(),
# MULTITHREADING
    'threads'           : set(),
    'workqueue'         : set(),
# MPI
    'mpi'               : set(),
    }



def main():
    if (len(sys.argv) != 2):
        sys.exit("Usage: techtree.py <.dep file>");

    validate_layout()
    Ma = expand_layout()   # Ma[m] is a set of allowable dependencies for module <m>, according to layout

    # Get the actual dependencies from the input file.
    # <D[m]> is a set of module dependencies for module <m>
    #  excluding esl_config.h and itself, as parsed out of input
    #  depfile. 
    #
    D = {}
    for line in open(sys.argv[1]):
        mat = re.match(r'(\S+): (.+)', line)  # First line in a list of dependencies?
        if mat:
            m  = process_token(mat.group(1))
            fields = mat.group(2).split()
            D[m] = set();
        else:
            fields   = line.split()

        for s in fields:
            if s != "\\" and s != 'esl_config.h':   # ignore \ and esl_config.h...
                s = process_token(s)                # ... extract module name from filename
                if s != m:                          # ... and ignore dependencies on itself
                    D[m].add(s)                     # ... otherwise, add dependencies to the set.
        
    output_direct_deplines(D)  # First section: actual dependencies, organized for study.
    compare_layout(Ma, D)      # Followed by any problems in them.


def process_token(tok):
    """
    """
    tok = re.sub(r'^esl_',   '', tok)
    tok = re.sub(r'.[coh]$', '', tok)
    return tok
####


# validate_layout()
#
# Checks the four global data structures that define the figure
# layout.
#
def validate_layout():
    Gl = set(easelgrps)             # <easelgrps> list is all the group names
    Gm = set(easelmods.keys())      # <easelmods> keys are also all the group names
    Gp = set(easelparents.keys())   # <easelparents> keys are also all the group names
    if Gl != Gm: sys.exit("easelgrps and easelmods have different set of group names")
    if Gl != Gp: sys.exit("easelgrps and easelparents have different set of group names")
    
    Mm = []            # Get all the module names in <easelmods>'s sets.
    for g in easelmods.keys():
        for m in easelmods[g]:
           Mm.append(m)
    if len(Mm) != len(set(Mm)):
        sys.exit("easelmods must have same module assigned to >1 group")
    Mm = set(Mm)
    Ma = set(easeladd.keys())
    if Mm != Ma:
        sys.exit("easelmods and easeladd have a different set of module names")

    print("# {0} groups".format(len(Gm)))
    print("# {0} modules".format(len(Mm)))
####


# expand_layout()
#
# Use the immediate group dependencies in <easelparents> to traverse
# upwards through the graph for each group <g>, to get the complete
# set of dependencies for it. Use that to obtain its complete set of
# module dependencies.
#
# Return a <Ma>, a dict of sets: Ma[m] is a set of modules that module
# <m> depends on.
#
def expand_layout():
    # Use a pushdown automaton to recursively find all group dependencies
    # for each group <g>; store this set in Ga[g].
    Ga = {}
    for g in easelparents.keys():
        Ga[g] = set()
        pda = list(easelparents[g])
        while len(pda):
            g2 = pda.pop()
            Ga[g].add(g2)
            pda.extend(list(easelparents[g2]))

    # Using Ga, plus additional within-group module dependencies in <adds>,
    # expand set of group dependencies to set of module dependencies for each module <m>.
    Ma = {}
    for g in easelmods.keys():
        mdep = set()
        for g2 in Ga[g]:                       # for each group that <g> depends on...
            mdep = mdep | set(easelmods[g2])   #   add the modules in that group
        for m in easelmods[g]:
            Ma[m] = mdep | easeladd[m]
    return Ma
####
    

def compare_layout(Ma, D):
    for m in Ma.keys():
        if not Ma[m] >= D[m]:
            print(m, D[m] - Ma[m])

def output_direct_deplines(D):
    for g in easelgrps:
        print('# {0}'.format(g))
        for m in easelmods[g]:
            print('{0:20s}: '.format(m), end='')

            for g2 in easelgrps:
                for m2 in easelmods[g2]:
                    if m2 in D[m]:
                        print('{0} '.format(m2), end='')
            print('')
    print('')

if __name__ == "__main__":
    main()
                
            

